% Cantilever Beam - Ov-IEFG-FEM (J.C.A.H & J.A.Z.C)
clear; clc;
%% Geometry and Properties
%Beam dimensions
LX=48; LY=12; % (m)
tol=1e-12;
%Mechanical (Isotropic Linear Elasticity) properties
E=3e7; %Young's Modulus (N/m^2)
nu=0.3; %Poisson Ration
PLTY=E*1e5;
P=-1000; %Total surface load (N)
I=LY^3/12; %Cross-Sectional Inertia Moment (m^3)
Ce=E/(1-nu^2)*[1 nu 0;nu 1 0;0 0 (1-nu)/2]; %Elastic moduli tensor in Voigt's notation (Plane Stress)
%% Discretization
%FEM Discretization
fac=6;
NX=24*2; NY=6*2;
dx=LX/NX; dy=LY/NY;
[xy,conn,nele,nnodes]=Malla_Quad_Uni(LX,LY,NX,NY);
C=zeros(nele,3,3);
C(:,1,1)=Ce(1,1); C(:,1,2)=Ce(1,2);
C(:,2,1)=Ce(2,1); C(:,2,2)=Ce(2,2); C(:,3,3)=Ce(3,3);
%IEFG Discretization
ss=5; 
ss2=7;
nx=4*fac;
dxOv=ss*dx/nx; ny=round(LY/dxOv); dyOv=LY/ny;
[xyOv,connOv,ncellsOv,nnodesOv]=Malla_Quad_Uni(ss*dx,LY,nx,ny);
eps=2.0; %Clustering parameter
xyiOv=xyOv;
 xyOv(:,1)=ss*dx*(1-tanh(eps*(ss*dx-xyiOv(:,1))/(ss*dx))/tanh(eps));
 xyOv(:,2)=LY/2+LY/2*tanh((2*xyiOv(:,2)-LY)*eps/LY)/tanh(eps);
%% Defining Influence Domains
%For IMLS-based reconstruction of FEM solution
dmax=2.0;
dm=dmax*[dx*ones(nnodes,1) dy*ones(nnodes,1)];
%For IMLS-based reconstruction of FEM solution
dmaxOv=2.0;
dmOv=dmaxOv*[dxOv*ones(nnodesOv,1) dyOv*ones(nnodesOv,1)];
 dmOv(:,1)=dmOv(:,1).*(eps/tanh(eps))./(cosh(eps*(ss*dx-xyiOv(:,1))/(ss*dx)).*cosh(eps*(ss*dx-xyiOv(:,1))/(ss*dx)));
 dmOv(:,2)=dmOv(:,2).*(eps/tanh(eps))./(cosh(eps*(2*xyiOv(:,2)-LY)/LY).*cosh(eps*(2*xyiOv(:,2)-LY)/LY));
xy(:,2)=xy(:,2)-LY/2;
xyOv(:,2)=xyOv(:,2)-LY/2;
%% Setting global Gauss integration points
quado=3; [gauss]=pgauss3(quado);
%For IMLS-based reconstruction of FEM solution
gs=egauss2D(xy,conn,gauss,nele);
numqc=nele*quado^2;
%For IEFG solution on overlapping nodes
gsOv=egauss2D(xyOv,connOv,gauss,ncellsOv);
numqcOv=ncellsOv*quado^2;
%% Construction of IMLS shape functions
%Background Nodes Shape functions computation (Time consuming part)
[PHI,DPHIX,DPHIY,DOM,Neig]=MLS_SHAPE_FUN(gs,xy,dm,numqc,nnodes);
DOMT=DOM;
Ext=DOMT==0; DOMT(Ext)=1;
%Overset Nodes Shape functions computation (Time consuming part)
[PHIOv,DPHIXOv,DPHIYOv,DOMOv,NeigOv]=MLS_SHAPE_FUN(gsOv,xyOv,dmOv,numqcOv,nnodesOv);
DOMOvT=DOMOv;
Ext=DOMOvT==0; DOMOvT(Ext)=1;
%% Assemble of global matrices and vectors
% FEM background model
[XGL,weight]=Gauss_FEM(gauss,quado);
GAUSS=struct('Loc_Coord',XGL,'weight',weight);
Km=STIFF_MATRIX(conn,xy,C,GAUSS); %Stiffness Matrix FEM
%Boundary for Dirichlet Condition
FIXNODES1=find(xy(:,1)<tol);
FIXNODES2=find(xy(:,1)>LX-tol);
DOF1=2*FIXNODES1-1; DOF2=2*FIXNODES1;
DOFd=[DOF1;DOF2];
Kp = sparse(DOFd,DOFd,PLTY,2*nnodes,2*nnodes); %Penalty Matrix for Dirichlet condition in FEM
Fp=zeros(2*nnodes,1);
Uax=-P*(2+nu)*xy(FIXNODES1,2)/(6*E*I).*(xy(FIXNODES1,2).^2-LY^2/4);
Uay=P*nu*LX/(2*E*I)*xy(FIXNODES1,2).^2;
Fp(DOF1)=Fp(DOF1)+Uax*PLTY;
Fp(DOF2)=Fp(DOF2)+Uay*PLTY;
%Boundary for Neumann Condition
DOF3=2*FIXNODES2;
Ft=zeros(2*nnodes,1); %Surface Loads Vector FEM
TRAC=P/(2*I)*(LY^2/4-xy(FIXNODES2,2).^2);
Ft(DOF3(1:end-1))=Ft(DOF3(1:end-1))+LY/(6*NY)*(2*TRAC(1:end-1)+TRAC(2:end));
Ft(DOF3(2:end))=Ft(DOF3(2:end))+LY/(6*NY)*(TRAC(1:end-1)+2*TRAC(2:end));
%----------------------------------------------------------------------------%
% IEFG Overset model
[~,BmatOv,~]=Shape_Vec(numqcOv,2,NeigOv,DPHIXOv,DPHIYOv,PHIOv);
COv=zeros(numqcOv,3,3);
COv(:,1,1)=Ce(1,1); COv(:,1,2)=Ce(1,2);
COv(:,2,1)=Ce(2,1); COv(:,2,2)=Ce(2,2); COv(:,3,3)=Ce(3,3);
KOv=STIFFNESS_MATRIX_2D(gsOv,numqcOv,NeigOv,BmatOv,DOMOv,COv,nnodesOv);  %Stiffness Matrix IEFG
%Boundary for Dirichlet Condition
l=xyOv(:,1)<tol; xyl=xyOv(l,:);
lthl=size(xyl,1);
conn1D=[1:1:lthl-1;2:1:lthl]; conn1D=conn1D';
[gsl]=egauss_Line_2D(xyl,conn1D,gauss);
[PHIlOv,DPHIXlOv,DPHIYlOv,DOMlOv,NeiglOv]=MLS_SHAPE_FUN(gsl,xyOv,dmOv,quado*(lthl-1),nnodesOv);
nn=zeros(quado*(lthl-1),2,2); nn(:,1,1)=1; nn(:,2,2)=1;
KpOv=PENALTY_MATRIX_VEC_2D(gsl,nnodesOv,PHIlOv,DOMlOv,nn,PLTY); %Penalty Matrix for Dirichlet condition in IEFG
UaxOv=-P*(2+nu)*gsl(:,2)/(6*E*I).*(gsl(:,2).^2-LY^2/4);
UayOv=P*nu*LX/(2*E*I)*gsl(:,2).^2;
n=zeros(quado*(lthl-1),2); n(:,1)=1; n(:,2)=1;
FpOv=PENALTY_VECTOR_VEC_2D(gsl,nnodesOv,PHIlOv,DOMlOv,n,PLTY,[UaxOv UayOv]);
%% Coupling Embedded Boundaries
%External immersed boundary to transfer information from FEM to IEFG
r=xyOv(:,1)>ss*dx-tol; xyr=xyOv(r,:);
lthr=size(xyr,1);
conn1D=[1:1:lthr-1;2:1:lthr]; conn1D=conn1D';
[gsr]=egauss_Line_2D(xyr,conn1D,gauss);
gsextOv=gsr;
nn=zeros(quado*(lthr-1),2,2); nn(:,1,1)=1; nn(:,2,2)=1;
[PHIextOv,DPHIXextOv,DPHIYextOv,DOMextOv,NeigextOv]=MLS_SHAPE_FUN(gsextOv,xyOv,dmOv,quado*(lthr-1),nnodesOv);
KpOvext=PENALTY_MATRIX_VEC_2D(gsextOv,nnodesOv,PHIextOv,DOMextOv,nn,PLTY); %Penalty Matrix for FEM to IEFG transfer
%Internal nodes below IEFG domain to transfer information from IEFG to FEM
in=find(xy(:,1)<ss/2*dx+tol&xy(:,1)>tol);
[PHIOvint,DPHIXOvint,DPHIYOvint,DOMOvint,NeigOvint]=MLS_SHAPE_FUN(xy(in,:),xyOv,dmOv,size(xy(in,:),1),nnodesOv);
DOMOvintT=DOMOvint;
Ext=DOMOvintT==0; DOMOvintT(Ext)=1;
DOF1=2*in-1; DOF2=2*in;
DOFin=[DOF1;DOF2];
Kpin=sparse(DOFin,DOFin,PLTY,2*nnodes,2*nnodes);  %Penalty Matrix for IEFG to FEM transfer
%% Solving Ov-IEFG-FEM
%Seed FEM solution
U=(Km+Kp)\(Ft);
Uxy=zeros(nnodes,2); Uxy(:,1)=U(1:2:end); Uxy(:,2)=U(2:2:end);
%Domain for local IMLS-based reconstruction of FEM solution
xye=(xy(conn(:,1),:)+xy(conn(:,2),:)+xy(conn(:,3),:)+xy(conn(:,4),:))/4;
box=xye(:,1)<ss2*dx+tol;
box_nodes=conn(box,:); box_nodes=box_nodes(:);
box_nodes=unique(box_nodes);
box=find(box);
prove = bsxfun(@plus, (box-1)*quado^2, 1:quado^2);
prove=prove(:);
gsbox=gs(prove,:);
xyp=xy(box_nodes,:); dmp=dm(box_nodes,:);
[PHIbox,DPHIboxX,DPHIboxY,DOMbox,Neigbox]=MLS_SHAPE_FUN(gsbox,xyp,dmp,size(gsbox,1),size(xyp,1)); %IMLS-based shape functions at Gauss Points in reconstruction domain
Mp=MASS_MATRIX(gsbox,size(gsbox,1),size(xyp,1),2,Neigbox,PHIbox,DOMbox,1); %Reconstruction matrix
[PHIp,DPHIpX,DPHIpY,DOMp,Neigp]=MLS_SHAPE_FUN(xyp,xyp,dmp,size(xyp,1),size(xyp,1)); %IMLS-based shape functions at nodes of elements forming the resconstruction domain
DOMpT=DOMp; Ext=DOMpT==0; DOMpT(Ext)=1;
LMBD=LAMBDA_MATRIX_Vec(PHIp,Neigp,DOMp); %Matrix for Lagrange Multiplier restrictions
D=[Mp LMBD;LMBD' sparse(2*size(xyp,1),2*size(xyp,1))]; %Reconstruction Matrix with Lagrange Multiplier restrictions
[PHIext,DPHIXext,DPHIYext,DOMext,Neigext]=MLS_SHAPE_FUN(gsextOv,xyp,dmp,size(gsextOv,1),size(xyp,1));
DOMextT=DOMext;
Ext=DOMextT==0; DOMextT(Ext)=1;
%Iterative Coupling
error=5; count=0;
tic;
while error>1e-8
count=count+1;
Ugx=Temp_Intp_FEM(conn,xy,GAUSS,Uxy(:,1));
Ugy=Temp_Intp_FEM(conn,xy,GAUSS,Uxy(:,2));
n=ones(size(gsbox,1),2);
Fpp=PENALTY_VECTOR_VEC_2D(gsbox,size(xyp,1),PHIbox,DOMbox,n,1,[Ugx(prove) Ugy(prove)]);
Up=zeros(2*size(xyp,1),1); Up(1:2:end)=Uxy(box_nodes,1); Up(2:2:end)=Uxy(box_nodes,2);
SOL=D\[Fpp;Up];
Un=SOL(1:2*size(xyp,1)); Unx=Un(1:2:end); Uny=Un(2:2:end);
UDOMextx=Unx(DOMextT);
Uextx=sum(PHIext.*UDOMextx,2);
UDOMexty=Uny(DOMextT);
Uexty=sum(PHIext.*UDOMexty,2);
n=ones(size(gsextOv,1),2);
FpOvext=PENALTY_VECTOR_VEC_2D(gsextOv,nnodesOv,PHIextOv,DOMextOv,n,PLTY,[Uextx Uexty]);
UOv=(KOv+KpOv+KpOvext)\(FpOvext+FpOv);
UOvx=UOv(1:2:end); UOvy=UOv(2:2:end);
UDOMOvintx=UOvx(DOMOvintT);
Uintx=sum(PHIOvint.*UDOMOvintx,2);
UDOMOvinty=UOvy(DOMOvintT);
Uinty=sum(PHIOvint.*UDOMOvinty,2);
if count>1
    errorint=sqrt((Uintx-Uintxp)'*(Uintx-Uintxp)+(Uinty-Uintyp)'*(Uinty-Uintyp))/sqrt(Uintx'*Uintx+Uinty'*Uinty);
    errorext=sqrt((Uextx-Uextxp)'*(Uextx-Uextxp)+(Uexty-Uextyp)'*(Uexty-Uextyp))/sqrt(Uextx'*Uextx+Uexty'*Uexty);
    error=errorint+errorext;
end
Fpin=zeros(2*nnodes,1); Fpin(2*in-1)=Uintx*PLTY; Fpin(2*in)=Uinty*PLTY;
U=(Km+Kp+Kpin)\(Ft+Fpin);
Uxy=zeros(nnodes,2); Uxy(:,1)=U(1:2:end); Uxy(:,2)=U(2:2:end);
Uintxp=Uintx; Uintyp=Uinty;
Uextxp=Uextx; Uextyp=Uexty;
end
toc;
display(error);
dUextxdx1=sum(DPHIXext.*UDOMextx,2);
DOMextOvT=DOMextOv;
Ext=DOMextOvT==0; DOMextOvT(Ext)=1;
UDOMOvextx=UOvx(DOMextOvT);
dUextxdx2=sum(DPHIXextOv.*UDOMOvextx,2);
UDOMpx=Unx(DOMpT);
Upx=sum(PHIp.*UDOMpx,2);
UDOMpy=Uny(DOMpT);
Upy=sum(PHIp.*UDOMpy,2);
[PHIsOv,DPHIXsOv,DPHIYsOv,DOMsOv,NeigsOv]=MLS_SHAPE_FUN(xyOv,xyOv,dmOv,nnodesOv,nnodesOv);
DOMsOvT=DOMsOv;
Ext=DOMsOvT==0; DOMsOvT(Ext)=1;
UsDOMOvx=UOvx(DOMsOvT);
UsOvx=sum(PHIsOv.*UsDOMOvx,2);
UsDOMOvy=UOvy(DOMsOvT);
UsOvy=sum(PHIsOv.*UsDOMOvy,2);
PlotMesh(xy+1000*Uxy,conn,'QUAD4');
axis equal;
hold on;
plot(xyOv(:,1)+1000*UsOvx,xyOv(:,2)+1000*UsOvy,'r.','MarkerSize',3);
CsOv=zeros(nnodesOv,3,3);
CsOv(:,1,1)=Ce(1,1); CsOv(:,1,2)=Ce(1,2);
CsOv(:,2,1)=Ce(2,1); CsOv(:,2,2)=Ce(2,2); CsOv(:,3,3)=Ce(3,3);
EPSsxxOv=sum(DPHIXsOv.*UsDOMOvx,2);
EPSsyyOv=sum(DPHIYsOv.*UsDOMOvy,2);
GAMsxyOv=sum(DPHIYsOv.*UsDOMOvx,2)+sum(DPHIXsOv.*UsDOMOvy,2);
SIGsxxOv=CsOv(:,1,1).*EPSsxxOv+CsOv(:,1,2).*EPSsyyOv;
SIGsyyOv=CsOv(:,2,1).*EPSsxxOv+CsOv(:,2,2).*EPSsyyOv;
TAUsxyOv=CsOv(:,3,3).*GAMsxyOv;
Mpg=MASS_MATRIX(gs,numqc,nnodes,2,Neig,PHI,DOM,1);
[PHIg,DPHIgX,DPHIgY,DOMg,Neigg]=MLS_SHAPE_FUN(xy,xy,dm,nnodes,nnodes);
DOMgT=DOMg; Ext=DOMgT==0; DOMgT(Ext)=1;
LMBDg=LAMBDA_MATRIX_Vec(PHIg,Neigg,DOMg);
Dg=[Mpg LMBDg;LMBDg' sparse(2*nnodes,2*nnodes)];
n=ones(numqc,2);
Fg=PENALTY_VECTOR_VEC_2D(gs,nnodes,PHI,DOM,n,1,[Ugx Ugy]);
SOLg=Dg\[Fg;U];
Ug=SOLg(1:2*nnodes); Ugx=Ug(1:2:end); Ugy=Ug(2:2:end);
UsDOMgx=Ugx(DOMgT);
UsDOMgy=Ugy(DOMgT);
Cn=zeros(nnodes,3,3);
Cn(:,1,1)=Ce(1,1); Cn(:,1,2)=Ce(1,2);
Cn(:,2,1)=Ce(2,1); Cn(:,2,2)=Ce(2,2); Cn(:,3,3)=Ce(3,3);
EPSgxx=sum(DPHIgX.*UsDOMgx,2);
EPSgyy=sum(DPHIgY.*UsDOMgy,2);
GAMgxy=sum(DPHIgY.*UsDOMgx,2)+sum(DPHIgX.*UsDOMgy,2);
SIGgxx=Cn(:,1,1).*EPSgxx+Cn(:,1,2).*EPSgyy;
SIGgyy=Cn(:,2,1).*EPSgxx+Cn(:,2,2).*EPSgyy;
TAUgxy=Cn(:,3,3).*GAMgxy;
[X,Y] = meshgrid(linspace(min(xy(:,1)),max(xy(:,1)),5*NX),linspace(min(xy(:,2)),max(xy(:,2)),5*NY));
[XOv,YOv] = meshgrid(linspace(min(xyOv(:,1)),max(xyOv(:,1)),5*nx),linspace(min(xyOv(:,2)),max(xyOv(:,2)),5*ny));
UX=griddata(xy(:,1),xy(:,2),Uxy(:,1),X,Y,'linear');
UXOv=griddata(xyOv(:,1),xyOv(:,2),full(UsOvx),XOv,YOv,'linear');
UY=griddata(xy(:,1),xy(:,2),Uxy(:,2),X,Y,'linear');
UYOv=griddata(xyOv(:,1),xyOv(:,2),full(UsOvy),XOv,YOv,'linear');
SIGXX=griddata(xy(:,1),xy(:,2),full(SIGgxx),X,Y,'linear');
SIGXXOv=griddata(xyOv(:,1),xyOv(:,2),full(SIGsxxOv),XOv,YOv,'linear');
SIGYY=griddata(xy(:,1),xy(:,2),full(SIGgyy),X,Y,'linear');
SIGYYOv=griddata(xyOv(:,1),xyOv(:,2),full(SIGsyyOv),XOv,YOv,'linear');
TAUXY=griddata(xy(:,1),xy(:,2),full(TAUgxy),X,Y,'linear');
TAUXYOv=griddata(xyOv(:,1),xyOv(:,2),full(TAUsxyOv),XOv,YOv,'linear');
a=X<ss*dx;
UX(a)=NaN;
UY(a)=NaN;
SIGXX(a)=NaN;
SIGYY(a)=NaN;
TAUXY(a)=NaN;
figure;
colormap('jet');
pcolor(X,Y,UY);shading interp; hold on;
pcolor(XOv,YOv,UYOv); shading interp;
plot(xyr(:,1),xyr(:,2),'w--','LineWidth',1);
axis equal; 
colorbar;
figure;
colormap('jet');
pcolor(X,Y,UX);shading interp; hold on;
pcolor(XOv,YOv,UXOv); shading interp;
plot(xyr(:,1),xyr(:,2),'w--','LineWidth',1);
axis equal; 
colorbar;
figure;
colormap('jet');
pcolor(X,Y,SIGXX);shading interp; hold on;
pcolor(XOv,YOv,SIGXXOv); shading interp;
plot(xyr(:,1),xyr(:,2),'w--','LineWidth',1);
axis equal; 
colorbar;
toc;